@@ -1,5 +1,5 @@
 /*
- *	$Id: HZfilter.c,v 3.5 1995/01/17 10:56:02 ygz Exp $
+ *	$Id: HZfilter.c,v 3.7 1995/08/06 07:26:34 ygz Exp $
  */
 
 /***********************************************************
@@ -56,7 +56,11 @@ static void DoPostSelection();
 
 int hzTableFilter (cxin, ch, strbuf)
     CXtermInputModule *cxin;
+#if NeedWidePrototypes
+    int ch;			/* the input keystroke */
+#else
     char ch;			/* the input keystroke */
+#endif
     char strbuf[];		/* the generated character(s) goes here */
 {
 #define CleanAssocDisplay()	\
@@ -247,6 +251,11 @@ int hzTableFilter (cxin, ch, strbuf)
 
 	} else {			/* in line editing mode */
 
+	    if (cxin->cursor == 0) {	/* already at the beginning */
+		Bell ();
+		return (0);
+	    }
+
 	    /* remove the key previous to the cursor */
 	    memmove(inbuf + cxin->cursor - 1, inbuf + cxin->cursor,
 		    cxin->hzinbufCount - cxin->cursor + 1);
@@ -459,9 +468,16 @@ static int AutoSegmentation(cxin, ch)
 {
   HZInputContext *ic = cxin->cur_ic;
 
-    /* only if we are at the end of the key buffer */
-    if ((!ic) || (! ic->im->m.do_auto_segment) || (cxin->cursor != -1))
-	return (0);
+    /* No auto segmentation if ... */
+    if ( (!ic) ||
+	/* ... we are not in any of these three type of ic ... */
+	((ic->type != ic_type_Trie) && (ic->type != ic_type_Trie_wc)
+		&& (ic->type != ic_type_Assoc)) ||
+	/* ... the input method doesn't allow so ... */
+	((! ic->im) || (! ic->im->m.do_auto_segment)) ||
+	/* ... we are not at the end of the key buffer. */
+	(cxin->cursor != -1) )
+	    return (0);
 
     /* insert an association key. */
     cxin->hzinbuf[ cxin->hzinbufCount ] = ic->im->k.def_assockey;
