@@ -384,7 +384,7 @@ void HZiaFlush(ia, screen)
     HZInputArea *ia;
     TScreen *screen;
 {
-#define DPY_ATTR_MASK	(INVERSE | UNDERLINE | BOLD | CHARDRAWN | CHARHANZI)
+#define DPY_ATTR_MASK	(INVERSE | UNDERLINE | BLINK | CHARDRAWN | CHARHANZI)
 
   int i;
 
@@ -457,7 +457,13 @@ void HZiaRedraw(ia, screen)
     HZInputArea *ia;
     TScreen *screen;
 {
-  int i;
+    int i;
+    extern XtermWidget term;
+
+    XSetForeground(screen->display, screen->hz_normalGC, screen->foreground);
+    XSetBackground(screen->display, screen->hz_normalGC, term->core.background_pixel);
+    XSetForeground(screen->display, screen->normalGC, screen->foreground);
+    XSetBackground(screen->display, screen->normalGC, term->core.background_pixel);
 
     HZiaClearText (screen, 0, 0, ia->rows, ia->cols - 1);
     HZiaDrawRuleLine (screen);
@@ -476,7 +482,14 @@ void HZiaShowMesg (ia, screen, mesgstr)
     TScreen *screen;
     char *mesgstr;
 {
-  register int col = 0;
+    register int col = 0;
+    extern XtermWidget term;
+
+    XSetForeground(screen->display, screen->hz_normalGC, screen->foreground);
+    XSetBackground(screen->display, screen->hz_normalGC, term->core.background_pixel);
+    XSetForeground(screen->display, screen->normalGC, screen->foreground);
+    XSetBackground(screen->display, screen->normalGC, term->core.background_pixel);
+
 
     while ((*mesgstr) && (col < ia->cols)) {
 	if ((*mesgstr) & 0x80)
@@ -612,9 +625,15 @@ static void HZiaDrawText (screen, row, col, str, nbytes, attr)
 	HZ_XDrawImString16 (screen->display, TextWindow(screen), gc,
 			screen->hz_normalFGCS, x, y, (XChar2b *)str, nbytes/2);
     } else {
+    char buf[128];
+    int i, len;
+    len = nbytes < 128 ? nbytes : 128;
+    for (i = 0; i < len; ++i) {
+        buf[i] = isprint (str[i]) ? str[i] : '-';
+    }
 	gc = (attr & INVERSE) ? screen->reverseGC : screen->normalGC;
 	XDrawImageString (screen->display, TextWindow(screen), gc,
-			x, y, (char *)str, nbytes);
+			x, y, buf, len);
     }
     if (attr & UNDERLINE)
 	XDrawLine(screen->display, TextWindow(screen), gc, x, y+1,
@@ -646,7 +665,8 @@ static void HZiaFillStippe (ia, screen)
     HZInputArea *ia;
     TScreen *screen;
 {
-  static GC gc = 0;
+    static GC gc = 0;
+    extern XtermWidget term;
 
     if (! gc) {		/* first time call, create a new one */
 	long fg = WhitePixel(screen->display, DefaultScreen(screen->display));
@@ -668,6 +688,9 @@ static void HZiaFillStippe (ia, screen)
     }
 
     XCopyGC(screen->display, screen->reverseGC, GCForeground|GCBackground, gc);
+    XSetForeground(screen->display, gc, term->core.background_pixel);
+    XSetBackground(screen->display, gc, term->core.background_pixel);
+
     XFillRectangle (screen->display, TextWindow(screen), gc,
 	screen->hzIwin.x, screen->hzIwin.y,
 	FontWidth (screen) * ia->cols, FontHeight(screen) * ia->rows);
