@@ -1,6 +1,6 @@
 #ifndef lint
-static char *cxterm_rid="$Id: main.c,v 3.2 1995/01/17 10:56:02 ygz Exp $";
-static char *rid="$XConsortium: main.c,v 1.222 94/04/17 20:23:28 gildea Exp $";
+static char *cxterm_rid="$Id: main.c,v 3.6 1995/08/09 04:45:03 ygz Exp $";
+static char *rid="$XConsortium: main.c,v 1.227.1.2 95/06/29 18:13:15 kaleb Exp $";
 #endif /* lint */
 
 /*
@@ -89,48 +89,44 @@ SOFTWARE.
 #include <pwd.h>
 #include <ctype.h>
 
+#define ARGUMENT_MAX_LENGTH 255
+
 #ifdef att
 #define ATT
 #endif
 
+#if defined(sgi) && (OSMAJORVERSION >= 5)
+#undef SYSV
+#ifndef SVR4
+#define SVR4
+#endif
+#endif
+
 #ifdef SVR4
+#ifndef SYSV
 #define SYSV			/* SVR4 is (approx) superset of SVR3 */
+#endif
 #define ATT
 #define USE_SYSV_UTMP
+/*#ifndef sgi*/			/* this doesn't really work, -ygz */ 
 #define USE_TERMIOS
+/*#endif*/
 #define HAS_UTMP_UT_HOST
 #endif
 
-#if defined(sgi) && OSMAJORVERSION >= 5
-#define SVR4			/* close enough for xterm */
-#define USE_SYSV_UTMP
-#define USE_TERMIOS
-#define HAS_UTMP_UT_HOST
-# ifdef CXTERM_COMPATIBLE
-#  if X11RELEASE == 5
-#   undef ATT			/* improperly turned on by IRIX's X11R5 */
-#  endif
-# endif
-#endif
-
-#ifdef SYSV386
+#if defined(SYSV) && defined(i386) && !defined(SVR4)
 #define USE_SYSV_UTMP
 #define ATT
 #define USE_HANDSHAKE
 static Bool IsPts = False;
 #endif
 
-#ifdef ATT
+#if defined(ATT) && !defined(sgi)
 #define USE_USG_PTYS
 #else
 #define USE_HANDSHAKE
 #endif
 
-#ifdef linux
-#define SYSV
-#define USE_SYSV_UTMP
-#endif
-
 #if defined(SYSV) && !defined(SVR4)
 /* older SYSV systems cannot ignore SIGHUP.
    Shell hangs, or you get extra shells, or something like that */
@@ -140,6 +136,18 @@ static Bool IsPts = False;
 #if defined(sony) && defined(bsd43) && !defined(KANJI)
 #define KANJI
 #endif
+
+#ifdef linux
+#define USE_SYSV_TERMIO
+#define USE_SYSV_PGRP
+#define USE_SYSV_UTMP
+#define USE_SYSV_SIGNALS
+#define HAS_UTMP_UT_HOST
+#define LASTLOG
+#define WTMP
+#include <sys/termios.h>
+#endif
+
 #include <sys/ioctl.h>
 #include <sys/stat.h>
 
@@ -203,9 +211,12 @@ static Bool IsPts = False;
 #define HAS_BSD_GROUPS
 #define USE_SYSV_UTMP
 #define HAS_UTMP_UT_HOST
+#define USE_POSIX_WAIT
 #include <sys/ptyio.h>
 #endif /* hpux */
 #ifdef sgi
+#define USE_SYSV_ENVVARS
+#define HAS_BSD_GROUPS
 #include <sys/sysmacros.h>
 #endif /* sgi */
 #ifdef sun
@@ -216,7 +227,9 @@ static Bool IsPts = False;
 #define HAS_UTMP_UT_HOST
 #endif
 #else /* } !SYSV { */			/* BSD systems */
+#ifndef linux
 #include <sgtty.h>
+#endif
 #include <sys/resource.h>
 #define HAS_UTMP_UT_HOST
 #define HAS_BSD_GROUPS
@@ -353,6 +366,10 @@ extern void exit();
 extern char *ttyname();
 #endif
 
+#ifdef sgi
+#include <locale.h>
+#endif
+
 #ifdef SYSV
 extern char *ptsname();
 #endif
@@ -377,6 +394,11 @@ static struct termio d_tio;
 #ifdef TIOCSLTC
 static struct ltchars d_ltc;
 #endif	/* TIOCSLTC */
+
+#ifdef sgi
+#undef TIOCLSET /* XXX why is this undef-ed again? */
+#endif
+
 #ifdef TIOCLSET
 static unsigned int d_lmode;
 #endif	/* TIOCLSET */
@@ -402,6 +424,41 @@ static struct jtchars d_jtc = {
 #endif /* sony */
 #endif /* USE_SYSV_TERMIO */
 
+/* allow use of system default characters if defined and reasonable */
+#ifndef CEOF
+#define CEOF ('D'&037)
+#endif
+#ifndef CSUSP
+#define CSUSP ('Z'&037)
+#endif
+#ifndef CQUIT
+#define CQUIT ('\\'&037)
+#endif
+#ifndef CEOL
+#define CEOL 0
+#endif
+#ifndef CSWTCH
+#define CSWTCH 0
+#endif
+#ifndef CLNEXT
+#define CLNEXT ('V'&037)
+#endif
+#ifndef CWERASE
+#define CWERASE ('W'&037)
+#endif
+#ifndef CRPRNT
+#define CRPRNT ('R'&037)
+#endif
+#ifndef CFLUSH
+#define CFLUSH ('O'&037)
+#endif
+#ifndef CSTOP
+#define CSTOP ('S'&037)
+#endif
+#ifndef CSTART
+#define CSTART ('Q'&037)
+#endif
+
 static int parse_tty_modes ();
 /*
  * SYSV has the termio.c_cc[V] and ltchars; BSD has tchars and ltchars;
@@ -450,7 +507,7 @@ struct _xttymodes {
 };
 
 #ifdef USE_SYSV_UTMP
-#if defined(X_NOT_STDC_ENV) || defined(AIXV3)
+#if defined(X_NOT_STDC_ENV) || (defined(AIXV3) && OSMAJORVERSION < 4)
 extern struct utmp *getutent();
 extern struct utmp *getutid();
 extern struct utmp *getutline();
@@ -523,6 +580,9 @@ static struct _resource {
     Boolean sunFunctionKeys;	/* %%% should be widget resource? */
     Boolean wait_for_map;
     Boolean useInsertMode;
+#ifdef sgi
+    Boolean useLocale;
+#endif
 } resource;
 
 /* used by VT (charproc.c) */
@@ -555,6 +615,10 @@ static XtResource application_resources[] = {
         offset(wait_for_map), XtRString, "false"},
     {"useInsertMode", "UseInsertMode", XtRBoolean, sizeof (Boolean),
         offset(useInsertMode), XtRString, "false"},
+#ifdef sgi
+    {"useLocale", "UseLocale", XtRBoolean, sizeof(Boolean),
+	offset(useLocale), XtRString, "true"},
+#endif
 };
 #undef offset
 
@@ -588,6 +652,8 @@ static XrmOptionDescRec optionDescList[] = {
 {"-cr",		"*cursorColor",	XrmoptionSepArg,	(caddr_t) NULL},
 {"-cu",		"*curses",	XrmoptionNoArg,		(caddr_t) "on"},
 {"+cu",		"*curses",	XrmoptionNoArg,		(caddr_t) "off"},
+{"-dc",		"*dynamicColors",XrmoptionNoArg,	(caddr_t) "off"},
+{"+dc",		"*dynamicColors",XrmoptionNoArg,	(caddr_t) "on"},
 {"-e",		NULL,		XrmoptionSkipLine,	(caddr_t) NULL},
 {"-fb",		"*boldFont",	XrmoptionSepArg,	(caddr_t) NULL},
 #ifdef HANZI
@@ -631,6 +697,10 @@ static XrmOptionDescRec optionDescList[] = {
 {"+t",		"*tekStartup",	XrmoptionNoArg,		(caddr_t) "off"},
 {"-tm",		"*ttyModes",	XrmoptionSepArg,	(caddr_t) NULL},
 {"-tn",		"*termName",	XrmoptionSepArg,	(caddr_t) NULL},
+#ifdef sgi
+{"-ul",		"*useLocale",	XrmoptionNoArg,		(caddr_t) "on"},
+{"+ul",		"*useLocale",	XrmoptionNoArg,		(caddr_t) "off"},
+#endif
 {"-ut",		"*utmpInhibit",	XrmoptionNoArg,		(caddr_t) "on"},
 {"+ut",		"*utmpInhibit",	XrmoptionNoArg,		(caddr_t) "off"},
 {"-im",		"*useInsertMode", XrmoptionNoArg,	(caddr_t) "on"},
@@ -656,6 +726,8 @@ static XrmOptionDescRec optionDescList[] = {
 {"-rv",		"*reverseVideo",XrmoptionNoArg,		(caddr_t) "on"},
 {"+rv",		"*reverseVideo",XrmoptionNoArg,		(caddr_t) "off"},
 {"-w",		".borderWidth", XrmoptionSepArg,	(caddr_t) NULL},
+{"-bbs",    "*bbs",         XrmoptionSepArg,    (XPointer)NULL},
+{"-usebold","*usebold",     XrmoptionSepArg,    (XPointer)NULL},
 };
 
 static struct _options {
@@ -666,6 +738,8 @@ static struct _options {
 { "-display displayname",  "X server to contact" },
 { "-geometry geom",        "size (in characters) and position" },
 { "-/+rv",                 "turn on/off reverse video" },
+{ "-bbs on/off",           "turn on/off ANSI-BBS emulation" },
+{ "-usebold on/off",       "Use bold/hilight for bold fonts"},
 { "-bg color",             "background color" },
 { "-fg color",             "foreground color" },
 { "-bd color",             "border color" },
@@ -718,6 +792,9 @@ static struct _options {
 { "-/+t",                  "turn on/off Tek emulation window" },
 { "-tm string",            "terminal mode keywords and characters" },
 { "-tn name",              "TERM environment variable name" },
+#ifdef sgi
+{ "-/+ul",                 "use/don't use locale for character input" },
+#endif
 #ifdef UTMP
 { "-/+ut",                 "turn on/off utmp inhibit" },
 #else
@@ -736,6 +813,7 @@ static struct _options {
 { "-C",                    "intercept console messages (not supported)" },
 #endif
 { "-Sxxd",                 "slave mode on \"ttyxx\", file descriptor \"d\"" },
+{ "-/+dc",		   "turn off/on dynamic color selection" },
 { NULL, NULL }};
 
 static char *message[] = {
@@ -879,10 +957,18 @@ char **argv;
 	register int pty;
 	int Xsocket, mode;
 	char *base_name();
-	int xerror(), xioerror();
+	int i, xerror(), xioerror();
 
 	ProgramName = argv[0];
 
+    for (i = 1; i < argc; ++i) {
+        if (strlen (argv[i]) > ARGUMENT_MAX_LENGTH) {
+            fprintf (stderr, "ABORT: Potential Security Violation !\n");
+            fprintf (stderr, "At least one of your arguments exceeds the length limit.\n");
+            exit (-1);
+        }
+    }
+            
 	ttydev = (char *) malloc (strlen (TTYDEV) + 1);
 	ptydev = (char *) malloc (strlen (PTYDEV) + 1);
 	if (!ttydev || !ptydev) {
@@ -894,25 +980,31 @@ char **argv;
 	strcpy (ttydev, TTYDEV);
 	strcpy (ptydev, PTYDEV);
 
-#ifdef USE_SYSV_TERMIO
+#ifdef USE_SYSV_TERMIO /* { */
 	/* Initialization is done here rather than above in order
 	** to prevent any assumptions about the order of the contents
 	** of the various terminal structures (which may change from
 	** implementation to implementation).
 	*/
-#if defined(macII) || defined(ATT) || defined(CRAY)
 	d_tio.c_iflag = ICRNL|IXON;
 	d_tio.c_oflag = OPOST|ONLCR|TAB3;
-#ifdef HANZI
-    	d_tio.c_cflag = B9600|CS8|CREAD|HUPCL;
-#else  /* HANZI */
+#if defined(macII) || defined(ATT) || defined(CRAY) || (defined(sun) && defined(SVR4)) /* { */
+#ifndef HANZI
     	d_tio.c_cflag = B9600|CS8|CREAD|PARENB|HUPCL;
+#else /* HANZI */
+    	d_tio.c_cflag = B9600|CS8|CREAD|HUPCL;	/* HANZI: cannot have PARENB */
 #endif /* HANZI */
     	d_tio.c_lflag = ISIG|ICANON|ECHO|ECHOE|ECHOK;
-
-#ifndef USE_TERMIOS
-	d_tio.c_line = 0;
+#ifdef ECHOKE
+	d_tio.c_lflag |= ECHOKE|IEXTEN;
 #endif
+#ifdef ECHOCTL
+	d_tio.c_lflag |= ECHOCTL|IEXTEN;
+#endif
+
+#ifndef USE_TERMIOS /* { */
+	d_tio.c_line = 0;
+#endif /* } */
 
 	d_tio.c_cc[VINTR] = CINTR;
 	d_tio.c_cc[VQUIT] = CQUIT;
@@ -923,54 +1015,94 @@ char **argv;
 	d_tio.c_cc[VEOL2] = CNUL;
 	d_tio.c_cc[VSWTCH] = CNUL;
 
-#ifdef USE_TERMIOS
+#ifdef USE_TERMIOS /* { */
 	d_tio.c_cc[VSUSP] = CSUSP;
 	d_tio.c_cc[VDSUSP] = CDSUSP;
-	d_tio.c_cc[VREPRINT] = CNUL;
-	d_tio.c_cc[VDISCARD] = CNUL;
-	d_tio.c_cc[VWERASE] = CNUL;
-	d_tio.c_cc[VLNEXT] = CNUL;
-#endif
-#ifdef TIOCSLTC
+	d_tio.c_cc[VREPRINT] = CRPRNT;
+	d_tio.c_cc[VDISCARD] = CFLUSH;
+	d_tio.c_cc[VWERASE] = CWERASE;
+	d_tio.c_cc[VLNEXT] = CLNEXT;
+#endif /* } */
+#ifdef TIOCSLTC /* { */
         d_ltc.t_suspc = CSUSP;		/* t_suspc */
         d_ltc.t_dsuspc = CDSUSP;	/* t_dsuspc */
-        d_ltc.t_rprntc = 0;		/* reserved...*/
-        d_ltc.t_flushc = 0;
-        d_ltc.t_werasc = 0;
-        d_ltc.t_lnextc = 0;
-#endif /* TIOCSLTC */
-#ifdef TIOCLSET
+        d_ltc.t_rprntc = CRPRNT;
+        d_ltc.t_flushc = CFLUSH;
+        d_ltc.t_werasc = CWERASE;
+        d_ltc.t_lnextc = CLNEXT;
+#endif /* } TIOCSLTC */
+#ifdef TIOCLSET /* { */
 	d_lmode = 0;
-#endif /* TIOCLSET */
-#else  /* else !macII */
-	d_tio.c_iflag = ICRNL|IXON;
-	d_tio.c_oflag = OPOST|ONLCR|TAB3;
-#ifdef BAUD_0
-#ifdef HANZI
-    	d_tio.c_cflag = CS8|CREAD|HUPCL;
-#else  /* HANZI */
+#endif /* } TIOCLSET */
+#else  /* }{ else !macII, ATT, CRAY */
+#ifdef BAUD_0 /* { */
+#ifndef HANZI
     	d_tio.c_cflag = CS8|CREAD|PARENB|HUPCL;
-#endif /* HANZI */
-#else	/* !BAUD_0 */
-#ifdef HANZI
-    	d_tio.c_cflag = B9600|CS8|CREAD|HUPCL;
-#else  /* HANZI */
+#else	/* HANZI */
+    	d_tio.c_cflag = CS8|CREAD|HUPCL;	/* HANZI: cannot have PARENB */
+#endif	/* HANZI */
+#else	/* }{ !BAUD_0 */
+#ifndef HANZI
     	d_tio.c_cflag = B9600|CS8|CREAD|PARENB|HUPCL;
-#endif /* HANZI */
-#endif	/* !BAUD_0 */
+#else	/* HANZI */
+    	d_tio.c_cflag = B9600|CS8|CREAD|HUPCL;	/* HANZI: cannot have PARENB */
+#endif	/* HANZI */
+#endif	/* } !BAUD_0 */
     	d_tio.c_lflag = ISIG|ICANON|ECHO|ECHOE|ECHOK;
-#ifndef sgi
-	d_tio.c_line = 0;
+#ifdef ECHOKE
+	d_tio.c_lflag |= ECHOKE|IEXTEN;
 #endif
-	d_tio.c_cc[VINTR] = 0x7f;		/* DEL  */
-	d_tio.c_cc[VQUIT] = '\\' & 0x3f;	/* '^\'	*/
-	d_tio.c_cc[VERASE] = '#';		/* '#'	*/
-	d_tio.c_cc[VKILL] = '@';		/* '@'	*/
-    	d_tio.c_cc[VEOF] = 'D' & 0x3f;		/* '^D'	*/
-	d_tio.c_cc[VEOL] = '@' & 0x3f;		/* '^@'	*/
+#ifdef ECHOCTL
+	d_tio.c_lflag |= ECHOCTL|IEXTEN;
+#endif
+#ifdef NTTYDISC
+        d_tio.c_line = NTTYDISC;
+#else
+	d_tio.c_line = 0;
+#endif	
+#ifdef sgi
+        d_tio.c_cflag &= ~(HUPCL|PARENB);
+        d_tio.c_iflag |= BRKINT|ISTRIP|IGNPAR;
+#endif
+	d_tio.c_cc[VINTR] = 'C' & 0x3f;		/* '^C'	*/
+	d_tio.c_cc[VERASE] = 0x7f;		/* DEL	*/
+	d_tio.c_cc[VKILL] = 'U' & 0x3f;		/* '^U'	*/
+	d_tio.c_cc[VQUIT] = CQUIT;		/* '^\'	*/
+    	d_tio.c_cc[VEOF] = CEOF;		/* '^D'	*/
+	d_tio.c_cc[VEOL] = CEOL;		/* '^@'	*/
 #ifdef VSWTCH
-	d_tio.c_cc[VSWTCH] = '@' & 0x3f;	/* '^@'	*/
-#endif	/* VSWTCH */
+	d_tio.c_cc[VSWTCH] = CSWTCH;            /* usually '^Z' */
+#endif
+#ifdef VLNEXT
+	d_tio.c_cc[VLNEXT] = CLNEXT;
+#endif
+#ifdef VWERASE
+	d_tio.c_cc[VWERASE] = CWERASE;
+#endif
+#ifdef VREPRINT
+	d_tio.c_cc[VREPRINT] = CRPRNT;
+#endif
+#ifdef VRPRNT
+	d_tio.c_cc[VRPRNT] = CRPRNT;
+#endif
+#ifdef VDISCARD
+	d_tio.c_cc[VDISCARD] = CFLUSH;
+#endif
+#ifdef VFLUSHO
+	d_tio.c_cc[VFLUSHO] = CFLUSH;
+#endif
+#ifdef VSTOP
+	d_tio.c_cc[VSTOP] = CSTOP;
+#endif
+#ifdef VSTART
+	d_tio.c_cc[VSTART] = CSTART;
+#endif
+#ifdef VSUSP
+	d_tio.c_cc[VSUSP] = CSUSP;
+#endif
+#ifdef VDSUSP
+	d_tio.c_cc[VDSUSP] = CDSUSP;
+#endif
 	/* now, try to inherit tty settings */
 	{
 	    int i;
@@ -986,36 +1118,69 @@ char **argv;
 		    d_tio.c_cc[VEOL] = deftio.c_cc[VEOL];
 #ifdef VSWTCH
 		    d_tio.c_cc[VSWTCH] = deftio.c_cc[VSWTCH];
-#endif /* VSWTCH */
+#endif
+#ifdef VEOL2
+		    d_tio.c_cc[VEOL2] = deftio.c_cc[VEOL2];
+#endif
+#ifdef VLNEXT
+		    d_tio.c_cc[VLNEXT] = deftio.c_cc[VLNEXT];
+#endif
+#ifdef VWERASE
+		    d_tio.c_cc[VWERASE] = deftio.c_cc[VWERASE];
+#endif
+#ifdef VREPRINT
+		    d_tio.c_cc[VREPRINT] = deftio.c_cc[VREPRINT];
+#endif
+#ifdef VRPRNT
+		    d_tio.c_cc[VRPRNT] = deftio.c_cc[VRPRNT];
+#endif
+#ifdef VDISCARD
+		    d_tio.c_cc[VDISCARD] = deftio.c_cc[VDISCARD];
+#endif
+#ifdef VFLUSHO
+		    d_tio.c_cc[VFLUSHO] = deftio.c_cc[VFLUSHO];
+#endif
+#ifdef VSTOP
+		    d_tio.c_cc[VSTOP] = deftio.c_cc[VSTOP];
+#endif
+#ifdef VSTART
+		    d_tio.c_cc[VSTART] = deftio.c_cc[VSTART];
+#endif
+#ifdef VSUSP
+		    d_tio.c_cc[VSUSP] = deftio.c_cc[VSUSP];
+#endif
+#ifdef VDSUSP
+		    d_tio.c_cc[VDSUSP] = deftio.c_cc[VDSUSP];
+#endif
 		    break;
 		}
 	    }
 	}
-#ifdef TIOCSLTC
+#ifdef TIOCSLTC /* { */
         d_ltc.t_suspc = '\000';		/* t_suspc */
         d_ltc.t_dsuspc = '\000';	/* t_dsuspc */
         d_ltc.t_rprntc = '\377';	/* reserved...*/
         d_ltc.t_flushc = '\377';
         d_ltc.t_werasc = '\377';
         d_ltc.t_lnextc = '\377';
-#endif	/* TIOCSLTC */
-#ifdef USE_TERMIOS
-	d_tio.c_cc[VSUSP] = '\000';
+#endif	/* } TIOCSLTC */
+#ifdef USE_TERMIOS /* { */
+	d_tio.c_cc[VSUSP] = CSUSP;
 	d_tio.c_cc[VDSUSP] = '\000';
 	d_tio.c_cc[VREPRINT] = '\377';
 	d_tio.c_cc[VDISCARD] = '\377';
 	d_tio.c_cc[VWERASE] = '\377';
 	d_tio.c_cc[VLNEXT] = '\377';
-#endif
-#ifdef TIOCLSET
+#endif /* } USE_TERMIOS */
+#ifdef TIOCLSET /* { */
 	d_lmode = 0;
-#endif	/* TIOCLSET */
-#endif  /* macII */
-#endif	/* USE_SYSV_TERMIO */
+#endif	/* } TIOCLSET */
+#endif  /* } macII, ATT, CRAY */
+#endif	/* } USE_SYSV_TERMIO */
 
 	/* Init the Toolkit. */
 	XtSetErrorHandler(xt_error);
-	toplevel = XtAppInitialize (&app_con, "XTerm", 
+	toplevel = XtAppInitialize (&app_con, "CXTerm", 
 				    optionDescList, XtNumber(optionDescList), 
 				    &argc, argv, fallback_resources, NULL, 0);
 
@@ -1023,6 +1188,11 @@ char **argv;
 				  application_resources,
 				  XtNumber(application_resources), NULL, 0);
 
+#ifdef sgi
+	if (resource.useLocale)
+	    setlocale(LC_ALL,"");
+#endif
+
 	waiting_for_initial_map = resource.wait_for_map;
 
 	/*
@@ -1071,7 +1241,6 @@ char **argv;
 	XtSetValues (toplevel, ourTopLevelShellArgs, 
 		     number_ourTopLevelShellArgs);
 
-
 	/* Parse the rest of the command line */
 	for (argc--, argv++ ; argc > 0 ; argc--, argv++) {
 	    if(**argv != '-') Syntax (*argv);
@@ -1082,6 +1251,7 @@ char **argv;
 		/* NOTREACHED */
 	     case 'C':
 #if defined(TIOCCONS) || defined(SRIOCSREDIR)
+#ifndef sgi
 		{
 		    struct stat sbuf;
 
@@ -1095,6 +1265,9 @@ char **argv;
 		    } else
 			Console = FALSE;
 		}
+#else /* sgi */
+		Console = TRUE;
+#endif /* sgi */
 #endif	/* TIOCCONS */
 		continue;
 	     case 'S':
@@ -1111,7 +1284,7 @@ char **argv;
 		if (argc <= 1) Syntax (*argv);
 		command_to_exec = ++argv;
 		break;
-	     default:
+        default:
 		Syntax (*argv);
 	    }
 	    break;
@@ -1128,6 +1301,22 @@ char **argv;
 
         screen = &term->screen;
 
+    /* This line is very rediculous, but I cannot find out the problem. :( */
+    screen->hilightcolor = term->misc.hilight;
+
+    if (term->misc.bbs) {
+        screen->ext_flags |= BBS;
+    }
+    else {
+        screen->ext_flags &= ~BBS;
+    }
+    if (term->misc.usebold) {
+        screen->ext_flags |= USEBOLD;
+    }
+    else {
+        screen->ext_flags &= ~USEBOLD;
+    }
+
 	if (screen->savelines < 0) screen->savelines = 0;
 
 	term->flags = 0;
@@ -1211,7 +1400,7 @@ char **argv;
 #endif /* HANZI */
 	}
 	if(i >= 0) {
-#if defined(USE_SYSV_TERMIO) && !defined(SVR4)
+#if defined(USE_SYSV_TERMIO) && !defined(SVR4) && !defined(linux)
 		/* SYSV has another pointer which should be part of the
 		** FILE structure but is actually a seperate array.
 		*/
@@ -1226,7 +1415,11 @@ char **argv;
 #endif
 		_bufend(stderr) = old_bufend;
 #else	/* USE_SYSV_TERMIO */
+#ifndef linux
 		stderr->_file = i;
+#else
+		setfileno(stderr, i);
+#endif
 #endif	/* USE_SYSV_TERMIO */
 
 		/* mark this file as close on exec */
@@ -1264,6 +1457,7 @@ char **argv;
 	screen->inhibit = inhibit;
 
 #ifdef AIXV3
+#if OSMAJORVERSION < 4
 	/* In AIXV3, xterms started from /dev/console have CLOCAL set.
 	 * This means we need to clear CLOCAL so that SIGHUP gets sent
 	 * to the slave-pty process when xterm exits. 
@@ -1281,6 +1475,7 @@ char **argv;
 		SysError(ERROR_TIOCSETP);
 	}
 #endif
+#endif
 #ifdef USE_SYSV_TERMIO
 	if (0 > (mode = fcntl(pty, F_GETFL, 0)))
 		Error();
@@ -1295,10 +1490,17 @@ char **argv;
 	mode = 1;
 	if (ioctl (pty, FIONBIO, (char *)&mode) == -1) SysError (ERROR_FIONBIO);
 #endif	/* USE_SYSV_TERMIO */
-	
-	pty_mask = 1 << pty;
-	X_mask = 1 << Xsocket;
-	Select_mask = pty_mask | X_mask;
+/* These several lines are updated by Zhuang Hao S.S.* */	
+    FD_ZERO (&pty_mask);
+    FD_ZERO (&X_mask);
+    FD_ZERO (&Select_mask);
+    FD_SET (pty, &pty_mask);
+    FD_SET (Xsocket, &X_mask);
+    FD_SET (pty, &Select_mask);
+    FD_SET (Xsocket, &Select_mask);
+    g__pty_fd = pty;
+    g__X_fd = Xsocket;
+
 	max_plus1 = (pty < Xsocket) ? (1 + Xsocket) : (1 + pty);
 
 #ifdef DEBUG
@@ -1333,7 +1535,7 @@ char *name;
 get_pty (pty)
     int *pty;
 {
-#if defined(SYSV) && defined(SYSV386)
+#if defined(SYSV) && defined(i386) && !defined(SVR4)
         /*
 	  The order of this code is *important*.  On SYSV/386 we want to open
 	  a /dev/ttyp? first if at all possible.  If none are available, then
@@ -1359,14 +1561,14 @@ get_pty (pty)
 	  */
         if (pty_search(pty) == 0)
 	    return 0;
-#endif /* SYSV && SYSV386 */
-#ifdef ATT
+#endif /* SYSV && i386 && !SVR4 */
+#if defined(ATT) && !defined(sgi)
 	if ((*pty = open ("/dev/ptmx", O_RDWR)) < 0) {
 	    return 1;
 	}
-#if defined(SVR4) || defined(SYSV386)
+#if defined(SVR4) || defined(i386)
 	strcpy(ttydev, ptsname(*pty));
-#if defined (SYSV) && defined(SYSV386)
+#if defined (SYSV) && defined(i386) && !defined(SVR4)
 	IsPts = True;
 #endif
 #endif
@@ -1559,6 +1761,12 @@ SIGNAL_T hungtty(i)
 	SIGNAL_RETURN;
 }
 
+/*
+ * declared outside USE_HANDSHAKE so HsSysError() callers can use
+ */
+static int pc_pipe[2];	/* this pipe is used for parent to child transfer */
+static int cp_pipe[2];	/* this pipe is used for child to parent transfer */
+
 #ifdef USE_HANDSHAKE
 typedef enum {		/* c == child, p == parent                        */
 	PTY_BAD,	/* c->p: can't open pty slave for some reason     */
@@ -1604,9 +1812,6 @@ int error;
 	exit(error);
 }
 
-static int pc_pipe[2];	/* this pipe is used for parent to child transfer */
-static int cp_pipe[2];	/* this pipe is used for child to parent transfer */
-
 void first_map_occurred ()
 {
     handshake_t handshake;
@@ -1624,11 +1829,19 @@ void first_map_occurred ()
 /*
  * temporary hack to get xterm working on att ptys
  */
+void
+HsSysError(pf, error)
+    int pf;
+    int error;
+{
+    fprintf(stderr, "%s: fatal pty error %d (errno=%d) on tty %s\n",
+	    xterm_name, error, errno, ttydev);
+    exit(error);
+}
 void first_map_occurred ()
 {
     return;
 }
-#define HsSysError(a,b)
 #endif /* USE_HANDSHAKE else !USE_HANDSHAKE */
 
 
@@ -1710,6 +1923,11 @@ spawn ()
 	screen->uid = getuid();
 	screen->gid = getgid();
 
+#ifdef linux
+	bzero(termcap, sizeof termcap);
+	bzero(newtc, sizeof newtc);
+#endif
+
 #ifdef SIGTTOU
 	/* so that TIOCSWINSZ || TIOCSIZE doesn't block */
 	signal(SIGTTOU,SIG_IGN);
@@ -1748,15 +1966,18 @@ spawn ()
 		 * Check results and ignore current control terminal if
 		 * necessary.  ENXIO is what is normally returned if there is
 		 * no controlling terminal, but some systems (e.g. SunOS 4.0)
-		 * seem to return EIO.
+		 * seem to return EIO.  Solaris 2.3 is said to return EINVAL.
 		 */
  		if (tty < 0) {
 			if (tty_got_hung || errno == ENXIO || errno == EIO ||
-			    errno == ENOTTY) {
+			    errno == EINVAL || errno == ENOTTY) {
 				no_dev_tty = TRUE;
 #ifdef TIOCSLTC
 				ltc = d_ltc;
 #endif	/* TIOCSLTC */
+#if defined(sun) && defined(SVR4)
+#undef TIOCLSET
+#endif
 #ifdef TIOCLSET
 				lmode = d_lmode;
 #endif	/* TIOCLSET */
@@ -1950,9 +2171,9 @@ spawn ()
 #endif
 
 #ifdef USE_USG_PTYS
-#if defined(SYSV) && defined(SYSV386)
+#if defined(SYSV) && defined(i386) && !defined(SVR4)
                 if (IsPts) {	/* SYSV386 supports both, which did we open? */
-#endif /* SYSV && SYSV386 */
+#endif /* SYSV && i386 && !SVR4 */
 		int ptyfd;
 
 		setpgrp();
@@ -1964,7 +2185,7 @@ spawn ()
 		if (ioctl (ptyfd, I_PUSH, "ptem") < 0) {
 		    SysError (2);
 		}
-#if !defined(SVR4) && !defined(SYSV386)
+#if !defined(SVR4) && !(defined(SYSV) && defined(i386))
 		if (!getenv("CONSEM") && ioctl (ptyfd, I_PUSH, "consem") < 0) {
 		    SysError (3);
 		}
@@ -1973,9 +2194,11 @@ spawn ()
 		    SysError (4);
 		}
 #ifdef SVR4			/* from Sony */
+# ifndef NO_TTCOMPAT		/* ! (sgi && IRIX 5.x) */
 		if (ioctl (ptyfd, I_PUSH, "ttcompat") < 0) {
 		    SysError (5);
 		}
+# endif
 #endif /* SVR4 */
 		tty = ptyfd;
 		close (screen->respond);
@@ -1993,9 +2216,9 @@ spawn ()
                         ws.ws_ypixel = FullHeight(screen);
                 }
 #endif
-#if defined(SYSV) && defined(SYSV386)
+#if defined(SYSV) && defined(i386) && !defined(SVR4)
                 } else {	/* else pty, not pts */
-#endif /* SYSV && SYSV386 */
+#endif /* SYSV && i386 && !SVR4 */
 #endif /* USE_USG_PTYS */
 
 #ifdef USE_HANDSHAKE		/* warning, goes for a long ways */
@@ -2087,7 +2310,7 @@ spawn ()
 
 			/* We have a new pty to try */
 			free(ttydev);
-			ttydev = malloc((unsigned)
+			ttydev = (char *)malloc((unsigned)
 			    (strlen(handshake.buffer) + 1));
 			strcpy(ttydev, handshake.buffer);
 		}
@@ -2098,12 +2321,12 @@ spawn ()
 		if (ptr = ttyname(tty))
 		{
 			/* it may be bigger */
-			ttydev = realloc (ttydev, (unsigned) (strlen(ptr) + 1));
+			ttydev = (char *)realloc (ttydev, (unsigned) (strlen(ptr) + 1));
 			(void) strcpy(ttydev, ptr);
 		}
-#if defined(SYSV) && defined(SYSV386)
+#if defined(SYSV) && defined(i386) && !defined(SVR4)
                 } /* end of IsPts else clause */
-#endif /* SYSV && SYSV386 */
+#endif /* SYSV && i386 && !SVR4 */
 
 #endif /* USE_HANDSHAKE -- from near fork */
 
@@ -2136,7 +2359,7 @@ spawn ()
 		 */
 		{
 #ifdef USE_SYSV_TERMIO
-#if defined(umips) || defined(CRAY)
+#if defined(umips) || defined(CRAY) || defined(linux) || (defined(sun) && defined(SVR4))
 		    /* If the control tty had its modes screwed around with,
 		       eg. by lineedit in the shell, or emacs, etc. then tio
 		       will have bad values.  Let's just get termio from the
@@ -2172,25 +2395,66 @@ spawn ()
 		    tio.c_cflag &= ~(CBAUD);
 		    tio.c_cflag |= B9600;
 #endif	/* !BAUD_0 */
-#ifdef HANZI
+		    tio.c_cflag &= ~CSIZE;
+#ifndef	HANZI
+		    if (screen->input_eight_bits)
+			tio.c_cflag |= CS8;
+		    else
+			tio.c_cflag |= CS7;
+#else	/* HANZI */
 		    /* set character size to 8 bits and disable parity */
 		    tio.c_cflag |= CS8;
 		    tio.c_cflag &= ~(PARENB);
-#endif /* HANZI */
+#endif	/* HANZI */
 		    /* enable signals, canonical processing (erase, kill, etc),
 		    ** echo
 		    */
-		    tio.c_lflag |= ISIG|ICANON|ECHO;
-		    /* reset EOL to defalult value */
-		    tio.c_cc[VEOL] = '@' & 0x3f;		/* '^@'	*/
+		    tio.c_lflag |= ISIG|ICANON|ECHO|ECHOE|ECHOK;
+#ifdef ECHOKE
+		    tio.c_lflag |= ECHOKE|IEXTEN;
+#endif
+#ifdef ECHOCTL
+		    tio.c_lflag |= ECHOCTL|IEXTEN;
+#endif
+		    /* reset EOL to default value */
+		    tio.c_cc[VEOL] = CEOL;			/* '^@' */
 		    /* certain shells (ksh & csh) change EOF as well */
-		    tio.c_cc[VEOF] = 'D' & 0x3f;		/* '^D'	*/
-#ifdef HANZI
-# ifdef   LPASS8
+		    tio.c_cc[VEOF] = CEOF;			/* '^D' */
+#ifdef VLNEXT
+		    tio.c_cc[VLNEXT] = CLNEXT;
+#endif
+#ifdef VWERASE
+		    tio.c_cc[VWERASE] = CWERASE;
+#endif
+#ifdef VREPRINT
+		    tio.c_cc[VREPRINT] = CRPRNT;
+#endif
+#ifdef VRPRNT
+		    tio.c_cc[VRPRNT] = CRPRNT;
+#endif
+#ifdef VDISCARD
+		    tio.c_cc[VDISCARD] = CFLUSH;
+#endif
+#ifdef VFLUSHO
+		    tio.c_cc[VFLUSHO] = CFLUSH;
+#endif
+#ifdef VSTOP
+		    tio.c_cc[VSTOP] = CSTOP;
+#endif
+#ifdef VSTART
+		    tio.c_cc[VSTART] = CSTART;
+#endif
+#ifdef VSUSP
+		    tio.c_cc[VSUSP] = CSUSP;
+#endif
+#ifdef VDSUSP
+		    tio.c_cc[VDSUSP] = CDSUSP;
+#endif
+#ifdef	HANZI
+# ifdef	LPASS8
 		    lmode |= LPASS8;
-# endif   /* LPASS8 */
-#endif /* HANZI */
-
+# endif	/* LPASS8 */
+#endif	/* HANZI */
 #define TMODE(ind,var) if (ttymodelist[ind].set) var = ttymodelist[ind].value;
 		    if (override_tty_modes) {
 			/* sysv-specific */
@@ -2201,7 +2465,37 @@ spawn ()
 			TMODE (XTTYMODE_eof, tio.c_cc[VEOF]);
 			TMODE (XTTYMODE_eol, tio.c_cc[VEOL]);
 #ifdef VSWTCH
-			TMODE (XTTYMODE_swtch, d_tio.c_cc[VSWTCH]);
+			TMODE (XTTYMODE_swtch, tio.c_cc[VSWTCH]);
+#endif
+#ifdef VSUSP
+			TMODE (XTTYMODE_susp, tio.c_cc[VSUSP]);
+#endif
+#ifdef VDSUSP
+			TMODE (XTTYMODE_dsusp, tio.c_cc[VDSUSP]);
+#endif
+#ifdef VREPRINT
+			TMODE (XTTYMODE_rprnt, tio.c_cc[VREPRINT]);
+#endif
+#ifdef VRPRNT
+			TMODE (XTTYMODE_rprnt, tio.c_cc[VRPRNT]);
+#endif
+#ifdef VDISCARD
+			TMODE (XTTYMODE_flush, tio.c_cc[VDISCARD]);
+#endif
+#ifdef VFLUSHO
+			TMODE (XTTYMODE_flush, tio.c_cc[VFLUSHO]);
+#endif
+#ifdef VWERASE
+			TMODE (XTTYMODE_weras, tio.c_cc[VWERASE]);
+#endif
+#ifdef VLNEXT
+			TMODE (XTTYMODE_lnext, tio.c_cc[VLNEXT]);
+#endif
+#ifdef VSTART
+			TMODE (XTTYMODE_start, tio.c_cc[VSTART]);
+#endif
+#ifdef VSTOP
+			TMODE (XTTYMODE_stop, tio.c_cc[VSTOP]);
 #endif
 #ifdef TIOCSLTC
 			/* both SYSV and BSD have ltchars */
@@ -2215,10 +2509,6 @@ spawn ()
 		    }
 #undef TMODE
 
-#ifndef HANZI
-		    if (ioctl (tty, TCSETA, &tio) == -1)
-			    HsSysError(cp_pipe[1], ERROR_TIOCSETP);
-#endif
 #ifdef TIOCSLTC
 		    if (ioctl (tty, TIOCSLTC, &ltc) == -1)
 			    HsSysError(cp_pipe[1], ERROR_TIOCSETC);
@@ -2227,11 +2517,8 @@ spawn ()
 		    if (ioctl (tty, TIOCLSET, (char *)&lmode) == -1)
 			    HsSysError(cp_pipe[1], ERROR_TIOCLSET);
 #endif	/* TIOCLSET */
-#ifdef HANZI
-		    /* change the order to avoid 8bit unset by lmode */
 		    if (ioctl (tty, TCSETA, &tio) == -1)
 			    HsSysError(cp_pipe[1], ERROR_TIOCSETP);
-#endif
 #else	/* USE_SYSV_TERMIO */
 		    sg.sg_flags &= ~(ALLDELAY | XTABS | CBREAK | RAW);
 		    sg.sg_flags |= ECHO | CRMOD;
@@ -2240,11 +2527,13 @@ spawn ()
 		    sg.sg_ospeed = B9600;
 		    /* reset t_brkc to default value */
 		    tc.t_brkc = -1;
-#ifdef sony
+#ifdef LPASS8
 		    if (screen->input_eight_bits)
 			lmode |= LPASS8;
 		    else
 			lmode &= ~(LPASS8);
+#endif
+#ifdef sony
 		    jmode &= ~KM_KANJI;
 #endif /* sony */
 #ifdef HANZI
@@ -2424,8 +2713,14 @@ spawn ()
 		(void) setutent ();
 		/* set up entry to search for */
 		ptyname = ttydev;
+#ifndef sgi
 		(void) strncpy(utmp.ut_id,ptyname + strlen(ptyname)-PTYCHARLEN,
 			       sizeof (utmp.ut_id));
+#else
+		(void) strncpy(utmp.ut_id,ptyname + sizeof("/dev/tty")-1,
+			       sizeof (utmp.ut_id));
+
+#endif
 		utmp.ut_type = DEAD_PROCESS;
 
 		/* position to entry in utmp file */
@@ -2505,7 +2800,8 @@ spawn ()
 					       XDisplayString (screen->display),
 					       sizeof(utmp.ut_host));
 #endif
-				time(&utmp.ut_time);
+				/* cast needed on Ultrix 4.4 */
+				time((Time_t*)&utmp.ut_time);
 				lseek(i, (long)(tslot * sizeof(struct utmp)), 0);
 				write(i, (char *)&utmp, sizeof(struct utmp));
 				close(i);
@@ -2687,7 +2983,7 @@ spawn ()
 			shname++;
 		else
 			shname = ptr;
-		shname_minus = malloc(strlen(shname) + 2);
+		shname_minus = (char *)malloc(strlen(shname) + 2);
 		(void) strcpy(shname_minus, "-");
 		(void) strcat(shname_minus, shname);
 #ifndef USE_SYSV_TERMIO
@@ -2769,7 +3065,7 @@ spawn ()
 			tslot = handshake.tty_slot;
 #endif	/* USE_SYSV_UTMP */
 			free(ttydev);
-			ttydev = malloc((unsigned) strlen(handshake.buffer) + 1);
+			ttydev = (char *)malloc((unsigned) strlen(handshake.buffer) + 1);
 			strcpy(ttydev, handshake.buffer);
 			break;
 		default:
@@ -3010,7 +3306,8 @@ nonblocking_wait()
 	/* cannot do non-blocking wait */
 	int pid = 0;
 #else	/* defined(USE_SYSV_SIGNALS) && (defined(CRAY) || !defined(SIGTSTP)) */
-	union wait status;
+/*	union wait status;*/
+	int status;		/* Modified by Zhuang Hao S.S.* for SCO */
 	register int pid;
 
 	pid = wait3 (&status, WNOHANG, (struct rusage *)NULL);
