@@ -177,7 +177,7 @@ typedef struct {
 typedef struct {
 	int		row;
 	int		col;
-	unsigned	flags;	/* Vt100 saves graphics rendition. Ugh! */
+	unsigned long	flags;	/* Vt100 saves graphics rendition. Ugh! */
 	char		curgl;
 	char		curgr;
 	char		gsets[4];
@@ -199,6 +199,72 @@ typedef struct {
 	int	linetype;
 } Tmodes;
 
+/***====================================================================***/
+
+#define	TEXT_FG		0
+#define	TEXT_BG		1
+#define	TEXT_CURSOR	2
+#define	MOUSE_FG	3
+#define	MOUSE_BG	4
+#define	TEK_FG		5
+#define	TEK_BG		6
+#define	NCOLORS		7
+
+#define	COLOR_DEFINED(s,w)	((s)->which&(1<<(w)))
+#define	COLOR_VALUE(s,w)	((s)->colors[w])
+#define	SET_COLOR_VALUE(s,w,v)	(((s)->colors[w]=(v)),((s)->which|=(1<<(w))))
+
+#define	COLOR_NAME(s,w)		((s)->names[w])
+#define	SET_COLOR_NAME(s,w,v)	(((s)->names[w]=(v)),((s)->which|=(1<<(w))))
+
+#define	UNDEFINE_COLOR(s,w)	((s)->which&=(~((w)<<1)))
+#define	OPPOSITE_COLOR(n)	(((n)==TEXT_FG?TEXT_BG:\
+				 ((n)==TEXT_BG?TEXT_FG:\
+				 ((n)==MOUSE_FG?MOUSE_BG:\
+				 ((n)==MOUSE_BG?MOUSE_FG:\
+				 ((n)==TEK_FG?TEK_BG:\
+				 ((n)==TEXT_BG?TEK_FG:(n))))))))
+
+typedef struct {
+	unsigned	which;
+	Pixel		colors[NCOLORS];
+	char		*names[NCOLORS];
+} ScrnColors;
+
+/***====================================================================***/
+
+#define MAXCOLORS 24
+#define COLORMASK 0x0F      /* Color Mask, only low 4 bits */
+#define	N_FGCOLOR 16
+#define N_COLOR   8
+#define BOLD_BIT        8   /* bit3 as bold-bit */
+
+#define FCOLOR_0		0
+#define FCOLOR_1		1
+#define FCOLOR_2		2
+#define FCOLOR_3		3
+#define FCOLOR_4		4
+#define FCOLOR_5		5
+#define FCOLOR_6		6
+#define FCOLOR_7		7
+#define FCOLOR_H0		8
+#define FCOLOR_H1		9
+#define FCOLOR_H2		10
+#define FCOLOR_H3		11
+#define FCOLOR_H4		12
+#define FCOLOR_H5		13
+#define FCOLOR_H6		14
+#define FCOLOR_H7		15
+
+#define BCOLOR_0		16
+#define BCOLOR_1		17
+#define BCOLOR_2		18
+#define BCOLOR_3		19
+#define BCOLOR_4		20
+#define BCOLOR_5		21
+#define BCOLOR_6		22
+#define BCOLOR_7		23
+
 typedef struct {
 	int Twidth;
 	int Theight;
@@ -246,10 +312,18 @@ typedef struct {
 	int		hz_normalFGCS;	/* normal painting	(GR/GL) */
 	int		hz_boldFGCS;	/* bold font painting	(GR/GL) */
 #endif /* HANZI */
-	Pixel		foreground;	/* foreground color		*/
-	Pixel		cursorcolor;	/* Cursor color			*/
-	Pixel		mousecolor;	/* Mouse color			*/
-	Pixel		mousecolorback;	/* Mouse color background	*/
+    /* These are variables added by Zhuang Hao S.S.* */
+    int             blink_state;
+    unsigned long   blink_on_ms;
+    unsigned long   blink_off_ms;
+    unsigned long   ext_flags;
+
+    Pixel       hilightcolor;   /* hilighted foreground color */
+    Pixel       foreground;	/* foreground color		*/
+    Pixel       cursorcolor;	/* Cursor color			*/
+    Pixel       mousecolor;	/* Mouse color			*/
+    Pixel       mousecolorback;	/* Mouse color background	*/
+    Pixel       colors[MAXCOLORS]; /* ANSI color emulation      */
 	int		border;		/* inner border			*/
 	Cursor		arrow;		/* arrow cursor			*/
 	unsigned short	send_mouse_pos;	/* user wants mouse transition  */
@@ -471,8 +545,12 @@ typedef struct _Misc {
     Boolean scrollbar;
     Boolean titeInhibit;
     Boolean tekSmall;	/* start tek window in small size */
+    Boolean dynamicColors;
     Boolean appcursorDefault;
     Boolean appkeypadDefault;
+    Boolean bbs;        /* if is in ANSI-BBS mode */
+    Boolean usebold;    /* if use bold fonts instead of hilight fonts */
+    Pixel   hilight;
 } Misc;
 
 typedef struct {int foo;} XtermClassPart, TekClassPart;
@@ -503,7 +581,9 @@ typedef struct _XtermWidgetRec {
     CorePart	core;
     TKeyboard	keyboard;	/* terminal keyboard		*/
     TScreen	screen;		/* terminal screen		*/
-    unsigned	flags;		/* mode flags			*/
+    unsigned long	flags;		/* mode flags			*/
+    unsigned    cur_foreground;	/* current foreground color	*/
+    unsigned    cur_background;	/* current background color	*/
     unsigned	initflags;	/* initial mode flags		*/
     Tabs	tabs;		/* tabstops of the terminal	*/
     Misc	misc;		/* miscellaneous parameters	*/
@@ -526,30 +606,29 @@ typedef struct _TekWidgetRec {
  */
 
 #ifdef HANZI
-#define	ATTRIBUTES	0x87	/* mask: user-visible attributes */
+#define	ATTRIBUTES	0xFF	/* mask: user-visible attributes */
 #else  /* HANZI */
 #define	ATTRIBUTES	0x07	/* mask: user-visible attributes */
 #endif /* HANZI */
 /* global flags and character flags (visible character attributes) */
 #define INVERSE		0x01	/* invert the characters to be output */
 #define UNDERLINE	0x02	/* true if underlining */
-#define BOLD		0x04
+#define BLINK		0x04
 /* character flags (internal attributes) */
-#define LINEWRAPPED	0x08	/* used on the first character in a
-				 * line to indicate that it wraps onto
-				 * the next line so we can tell the
-				 * difference between lines that have
-				 * wrapped around and lines that have
-				 * ended naturally with a CR at column
-				 * max_col.
-				 */
-#define CHARDRAWN	0x10    /* a character has been drawn here on the
+#define CHARDRAWN	0x08	 /* a character has been drawn here on the
 				   screen.  Used to distinguish blanks from
 				   empty parts of the screen when selecting */
+
+#define FG_COLOR        0x10     /* true if background set */
+#define BG_COLOR        0x20     /* true if foreground set */
+
 #ifdef HANZI
-#define FIRSTBYTE	0x40	/* is the first byte in a Hanzi (internal) */
-#define CHARHANZI	0x80	/* is part of a Hanzi   (global & visible) */
+#define FIRSTBYTE	0x80	/* is the first byte in a Hanzi (internal) */
+#define CHARHANZI	0x40	/* is part of a Hanzi   (global & visible) */
 #endif /* HANZI */
+/* This attribute is NOT written into buf[3], only in term->flags */
+#define BOLD   0x100   /* The color is hilighted */
+
 /* global flags */
 #define WRAPAROUND	0x400	/* true if auto wraparound mode */
 #define	REVERSEWRAP	0x800	/* true if reverse wraparound mode */
@@ -561,6 +640,10 @@ typedef struct _TekWidgetRec {
 #define IN132COLUMNS	0x20000	/* true if in 132 column mode */
 
 
+/* Bit masks for newly added ext_flags, by Zhuang Hao S.S.* */
+#define BBS         1       /* ANSI-BBS Mode */
+#define USEBOLD     2       /* Use bold fonts */
+
 #define VWindow(screen)		(screen->fullVwin.window)
 #define VShellWindow		term->core.parent->core.window
 #define TextWindow(screen)      (screen->fullVwin.window)
@@ -599,7 +682,7 @@ typedef struct _TekWidgetRec {
 # define ROWSINPUTAREA	2	/* number of rows in hanzi input areas */
 
 # define FGCSinScreen(s,f)	\
-		(((f) & BOLD) ? (s)->hz_boldFGCS : (s)->hz_normalFGCS)
+		(((f) & BOLD_BIT) ? (s)->hz_boldFGCS : (s)->hz_normalFGCS)
 #endif /* HANZI */
 
 #define TEK_LINK_BLOCK_SIZE 1024
