@@ -3,14 +3,17 @@
 
 # definitions
 
-: ${ECHONONL:="echo -n"}
+if [ "$ECHONONL" = "" ]; then
+    ECHONONL="echo -n"
+fi
 
 # add CXterm.ins to .Xdefaults ...
 
 if [ -f ./cxterm.ins ]; then
     . ./cxterm.ins
 else
-    CXTERM_AD=cxterm/CXterm.ad
+    echo "Cannot find installation record.  Please install cxterm first."
+    exit 1
 fi
 
 if   [ -f $HOME/.Xdefaults ]; then
@@ -23,6 +26,8 @@ else
     X_RES=$HOME/.Xdefaults
 fi
 
+CXTERM_AD=./cxterm/CXterm.ad
+
 echo "I am putting $CXTERM_AD into your X resource file"
 echo "$X_RES.  If this is a wrong place, please make the adjustment"
 echo "yourself.  Refer to Doc/x-resrc file for more information."
@@ -49,3 +54,4 @@ echo ""
 echo "Finally, remember to add the directory where you installed cxterm,"
 echo "$BINDIR, to your PATH (in .cshrc or .profile, etc.)"
 
+exit 0
