@@ -1,5 +1,5 @@
 /*
- *	$Id: tit2cit.c,v 2.2 1995/01/17 10:59:07 ygz Exp $
+ *	$Id: tit2cit.c,v 2.3 1995/02/14 05:52:43 ygz Exp $
  */
 
 /***********************************************************************
@@ -798,6 +798,13 @@ static void ParseKey (buf)
     register char *p1 = buf, *p2 = buf;
 
     while (*p1) {
+	if (*p1 & 0x80) {	/* a hanzi, skip it */
+	    *p2++ = *p1++;
+	    if (! *p1)
+		break;
+	    *p2++ = *p1++;
+	    continue;
+	}
 	if (*p1 == '\\') {
 	    p1++;
 	    if ((p1[0] == '0') || (p1[0] == '1')) {
